class DeviceProxy:
    def ScanNumber(self):
        pass

    def State(self):
        pass

    def SequenceScanNumber(self):
        pass

    def SavingFile(self):
        pass

    def ScanTitle(self):
        pass
